/* X11Helper.c --  This "helper" process, forked off by the magic process,
 *   looks like a kludge (and is), but is the only way to ensure that X11
 *   is able to run its event-driven protocol (through calls to XNextEvent)
 *   while magic runs its interrupt-driven protocol (through calls to
 *   select()).  Both are infinite (blocking) loops and must execute in
 *   parallel.  Xlib is not thread-safe, so running this as a thread
 *   instead of a forked process is a bad idea (confuses X11 and locks up).
 */

#include <stdio.h>
#include <signal.h>
#include <X11/Xlib.h>
#include <X11/keysym.h>
#include <X11/Xutil.h>

/* 
 * Portability stuff
 */

#if     (defined(MIPSEB) && defined(SYSTYPE_BSD43)) || ibm032
# define        SIG_RETURNS_INT
#endif

/* Some machines have signal handlers returning an int, while other machines
 * have it returning a void.  If you have a machine that requires ints put 
 * it in the list of machines in misc/magic.h.
 */
#ifdef	SIG_RETURNS_INT
#define	sigRetVal	int
#else
#define	sigRetVal	void
#endif

/*
 * Declaration of global variables and procedure prototypes
 */

int readPipe,writePipe;	/* pipe file descriptor to magic process */
int parentID;		/* process id of parent */
Display *grXdpy;	/* X11 display */

sigRetVal MapWindow();

/*
 * Main program:
 *   This infinite loop is a rewrite of XtMainLoop() which includes both
 *   the ability to handle interrupts and communicate with magic's main
 *   process via the I/O pipe.  The program exits when killed by the
 *   GrClose routine in magic.
 */

main (argc, argv)
    int argc;
    char **argv;
{
    XEvent xevent;

#ifdef macII
    set42sig();
#endif macII

    if (argc <= 1) {
	fprintf(stderr, "%s: expecting two file-descriptor numbers.\n\
	    \t(This program should be run only by magic itself.)\n", argv[0]);
	exit(1);
    }

    sscanf(argv[1], "%d %d", &readPipe,&writePipe);
    grXdpy = XOpenDisplay(0);
    if (grXdpy == (Display *) 0) {
	fprintf(stderr,"%s: XOpenDisplay failed.\n", argv[0]);
	exit(1);
    }

    parentID = getppid();

    signal(SIGINT,SIG_IGN);
    signal(SIGQUIT,SIG_IGN);
    signal(SIGTERM,MapWindow); 
#ifdef SIGTSTP
    signal(SIGTSTP,SIG_IGN);
#endif
#ifdef SIGCONT
    signal(SIGCONT,SIG_IGN);
#endif

    while (1)
    {
	XNextEvent(grXdpy, &xevent);
	ParseEvent(&xevent);
    }
}

/*
 * ParseEvent(XEvent *event):
 *   Pass XEvents on to the magic main process by writing into the
 *   connecting pipe.  Keystrokes are handled by translation from a
 *   key code to an X11 KeySym, with the KeySym string passed on to
 *   magic through the I/O pipe.
 */

ParseEvent (event)
    XEvent *event;
{
    if (event->type == KeyPress) 
    {
	XKeyPressedEvent *KeyPressedEvent = (XKeyPressedEvent *) event;
	char inChar[10],c,*p;
	int nbytes;
	KeySym keysym;

	nbytes = XLookupString(KeyPressedEvent, inChar, sizeof(inChar),
			 &keysym, NULL);
	if (nbytes == 0)
	{
	    if (!IsModifierKey(keysym))
	    {
		if (keysym == XK_Left)
		{
		    c = '\002'; /* Ctrl-b */
		}
		else if (keysym == XK_Right)
		{
		    c = '\006'; /* Ctrl-f */
		}
		else if (keysym == XK_Up)
		{
		    c = '\020'; /* Ctrl-p */
		}
		else if (keysym == XK_Down)
		{
		    c = '\016'; /* Ctrl-n */
		}
		else if (keysym == XK_BackSpace)
		{
		    c = '\010'; /* Ctrl-h */
		}
		else if (keysym == XK_Delete)
		{
		    c = '\010'; /* Ctrl-h */
		}
		else if (keysym == XK_Home)
		{
		    c = '\001'; /* Ctrl-a */
		}
		else if (keysym == XK_End)
		{
		    c = '\005'; /* Ctrl-e */
		}
		else
		{
		    c = '\0';
		}
		write(writePipe, event, sizeof(XEvent));
		write(writePipe, &c, sizeof(char));
#ifndef USE_IO_PROBE
		kill(parentID, SIGIO);
#endif
	    }
	}
	else
	{
	    p = inChar;
	    while (nbytes--)
	    {
		if ( (c = *p++) == 3)
		{
		    kill(parentID, SIGINT);
		} 
		else
		{
		    write(writePipe, event, sizeof(XEvent));
		    write(writePipe, &c, sizeof(char));
#ifndef USE_IO_PROBE
		    kill(parentID, SIGIO);
#endif
		}
	    } 
	}
    }
    else  /* All event types other than KeyPress */
    {
	write(writePipe, event, sizeof(XEvent));
#ifndef USE_IO_PROBE
	kill(parentID, SIGIO);
#endif
    }
}

/*
 * MapWindow():
 *   On startup of any new magic window, magic writes the X11 window ID
 *   into the communication pipe and manually generates a SIGTERM signal.
 *   The interrupt handler calls MapWindow() on receiving SIGTERM, and
 *   the window ID is read from the I/O pipe.  MapWindow() then activates
 *   the window by declaring what events it will interpret for that
 *   window (keystrokes, mouse button, and expose and resize events)
 */

sigRetVal
MapWindow()
{
    Window window;

    if ( read(readPipe, (char *)&window, sizeof(Window)) == sizeof(Window))
    {
	XSelectInput(grXdpy, window,
		     KeyPressMask|ButtonPressMask|
		     ButtonReleaseMask|ExposureMask|
		     StructureNotifyMask|
		     OwnerGrabButtonMask);
	XSync(grXdpy,1);

        /* BSD keeps the signal handler.  SYSV destroys it, so it must	*/
	/* be re-declared at the end of the interrupt routine.		*/

#ifdef SYSV
        signal(SIGTERM, MapWindow);
#endif SYSV
    }
    else
	fprintf(stderr,"X11Helper: read on pipe failed\n");
}
